"""A customizable simple initramfs generator"""

__author__ = 'Louis Leseur'
__copyright__ = 'Copyright (c) 2020 Louis Leseur'
__email__ = 'louis.leseur@gmail.com'
__license__ = 'MIT'
__url__ = 'https://github.com/teapot9/fand'
__version__ = '0.2.0'

import logging

logging.basicConfig(
    format='%(levelname)s: %(message)s',
    level=logging.NOTSET,
)
